from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union, Callable, Any, List

import requests
from msal import (
    ConfidentialClientApplication,
    ManagedIdentityClient,
    UserAssignedManagedIdentity,
    SystemAssignedManagedIdentity
)
from redis.auth.err import RequestTokenErr
from redis.auth.idp import IdentityProviderInterface
from redis.auth.token import TokenInterface, JWToken


class ManagedIdentityType(Enum):
    USER_ASSIGNED = UserAssignedManagedIdentity
    SYSTEM_ASSIGNED = SystemAssignedManagedIdentity


class ManagedIdentityIdType(Enum):
    CLIENT_ID = "client_id"
    OBJECT_ID = "object_id"
    RESOURCE_ID = "resource_id"


@dataclass
class ManagedIdentityProviderConfig:
    identity_type: ManagedIdentityType
    resource: str
    id_type: Optional[ManagedIdentityIdType] = None
    id_value: Optional[str] = ''
    kwargs: Optional[dict] = field(default_factory=dict)


@dataclass
class ServicePrincipalIdentityProviderConfig:
    client_credential: Any
    client_id: str
    scopes: Optional[List[str]] = None
    timeout: Optional[float] = None
    tenant_id: Optional[str] = None
    token_kwargs: Optional[dict] = None
    app_kwargs: Optional[dict] = field(default_factory=dict)


class EntraIDIdentityProvider(IdentityProviderInterface):
    """
    EntraID Identity Provider implementation.
    It's recommended to use an additional factory methods to simplify object instantiation.

    Methods: create_provider_from_managed_identity, create_provider_from_service_principal.
    """
    def __init__(
            self,
            app: Union[ManagedIdentityClient, ConfidentialClientApplication],
            scopes : List = [],
            resource: str = '',
            **kwargs
    ):
        self._app = app
        self._scopes = scopes
        self._resource = resource
        self._kwargs = kwargs

    def request_token(self, force_refresh=False) -> TokenInterface:
        """
        Request token from identity provider.
        Force refresh argument is optional and works only with Service Principal auth method.

        :param force_refresh:
        :return: TokenInterface
        """
        if isinstance(self._app, ManagedIdentityClient):
            return self._get_token(self._app.acquire_token_for_client, resource=self._resource)

        if force_refresh:
            self._app.remove_tokens_for_client()

        return self._get_token(
                self._app.acquire_token_for_client,
                scopes=self._scopes,
                **self._kwargs
        )

    def _get_token(self, callback: Callable, **kwargs) -> JWToken:
        try:
            response = callback(**kwargs)

            if "error" in response:
                raise RequestTokenErr(response["error_description"])

            return JWToken(callback(**kwargs)["access_token"])
        except Exception as e:
            raise RequestTokenErr(e)


def _create_provider_from_managed_identity(config: ManagedIdentityProviderConfig) -> EntraIDIdentityProvider:
    """
    Create an EntraID identity provider following Managed Identity auth flow.

    :param config: Config for managed assigned identity provider
    See: :class:`ManagedIdentityClient` acquire_token_for_client method.

    :return: :class:`EntraIDIdentityProvider`
    """
    if config.identity_type == ManagedIdentityType.USER_ASSIGNED:
        if config.id_type is None or config.id_value == '':
            raise ValueError("Id_type and id_value are required for User Assigned identity auth")

        kwargs = {
            config.id_type.value: config.id_value
        }

        managed_identity = config.identity_type.value(**kwargs)
    else:
        managed_identity = config.identity_type.value()

    app = ManagedIdentityClient(managed_identity, http_client=requests.Session())
    return EntraIDIdentityProvider(app, [], config.resource, **config.kwargs)


def _create_provider_from_service_principal(config: ServicePrincipalIdentityProviderConfig) -> EntraIDIdentityProvider:
    """
    Create an EntraID identity provider following Service Principal auth flow.

    :param config: Config for service principal identity provider

    :return: :class:`EntraIDIdentityProvider`
    See: :class:`ConfidentialClientApplication`.
    """

    if config.scopes is None:
        scopes = ["https://redis.azure.com/.default"]
    else:
        scopes = config.scopes

    authority = f"https://login.microsoftonline.com/{config.tenant_id}" \
        if config.tenant_id is not None else config.tenant_id

    app = ConfidentialClientApplication(
        client_id=config.client_id,
        client_credential=config.client_credential,
        timeout=config.timeout,
        authority=authority,
        **config.app_kwargs
    )
    return EntraIDIdentityProvider(app, scopes, **config.token_kwargs)



from typing import Union, Tuple, Callable, Any, Awaitable, Optional, List

from redis.credentials import StreamingCredentialProvider
from redis.auth.token_manager import TokenManagerConfig, RetryPolicy, TokenManager, CredentialsListener

from redis_entraid.identity_provider import ManagedIdentityType, ManagedIdentityIdType, \
    _create_provider_from_managed_identity, ManagedIdentityProviderConfig, ServicePrincipalIdentityProviderConfig, \
    _create_provider_from_service_principal

DEFAULT_EXPIRATION_REFRESH_RATIO = 0.7
DEFAULT_LOWER_REFRESH_BOUND_MILLIS = 0
DEFAULT_TOKEN_REQUEST_EXECUTION_TIMEOUT_IN_MS = 100
DEFAULT_MAX_ATTEMPTS = 3
DEFAULT_DELAY_IN_MS = 3

class EntraIdCredentialsProvider(StreamingCredentialProvider):
    def __init__(
            self,
            idp_config: Union[ManagedIdentityProviderConfig, ServicePrincipalIdentityProviderConfig],
            token_manager_config: TokenManagerConfig,
            initial_delay_in_ms: float = 0,
            block_for_initial: bool = False,
    ):
        """
        :param idp_config: Identity provider specific configuration.
        :param token_manager_config: Token manager specific configuration.
        :param initial_delay_in_ms: Initial delay before run background refresh (valid for async only)
        :param block_for_initial: Block execution until initial token will be acquired (valid for async only)
        """
        if isinstance(idp_config, ManagedIdentityProviderConfig):
            idp = _create_provider_from_managed_identity(idp_config)
        else:
            idp = _create_provider_from_service_principal(idp_config)

        self._token_mgr = TokenManager(
            idp,
            token_manager_config
        )
        self._listener = CredentialsListener()
        self._is_streaming = False
        self._initial_delay_in_ms = initial_delay_in_ms
        self._block_for_initial = block_for_initial

    def get_credentials(self) -> Union[Tuple[str], Tuple[str, str]]:
        """
        Acquire token from the identity provider.
        """
        init_token = self._token_mgr.acquire_token()

        if self._is_streaming is False:
            self._token_mgr.start(
                self._listener,
                skip_initial=True
            )
            self._is_streaming = True

        return init_token.get_token().try_get('oid'), init_token.get_token().get_value()

    async def get_credentials_async(self) -> Union[Tuple[str], Tuple[str, str]]:
        """
        Acquire token from the identity provider in async mode.
        """
        init_token = await self._token_mgr.acquire_token_async()

        if self._is_streaming is False:
            await self._token_mgr.start_async(
                self._listener,
                initial_delay_in_ms=self._initial_delay_in_ms,
                block_for_initial=self._block_for_initial,
                skip_initial=True
            )
            self._is_streaming = True

        return init_token.get_token().try_get('oid'), init_token.get_token().get_value()

    def on_next(self, callback: Union[Callable[[Any], None], Awaitable]):
        self._listener.on_next = callback

    def on_error(self, callback: Union[Callable[[Exception], None], Awaitable]):
        self._listener.on_error = callback

    def is_streaming(self) -> bool:
        return self._is_streaming


def create_from_managed_identity(
        identity_type: ManagedIdentityType,
        resource: str,
        id_type: Optional[ManagedIdentityIdType] = None,
        id_value: Optional[str] = '',
        kwargs: Optional[dict] = {},
        token_manager_config: Optional[TokenManagerConfig] = TokenManagerConfig(
            DEFAULT_EXPIRATION_REFRESH_RATIO,
            DEFAULT_LOWER_REFRESH_BOUND_MILLIS,
            DEFAULT_TOKEN_REQUEST_EXECUTION_TIMEOUT_IN_MS,
            RetryPolicy(
                DEFAULT_MAX_ATTEMPTS,
                DEFAULT_DELAY_IN_MS
            )
        )
) -> EntraIdCredentialsProvider:
    """
    Create a credential provider from a managed identity type.

    :param identity_type: Managed identity type.
    :param resource: Identity provider resource.
    :param id_type: Identity provider type.
    :param id_value: Identity provider value.
    :param kwargs: Optional keyword arguments to pass to identity provider. See: :class:`ManagedIdentityClient`
    :param token_manager_config: Token manager specific configuration.
    :return: EntraIdCredentialsProvider instance.
    """
    managed_identity_config = ManagedIdentityProviderConfig(
        identity_type=identity_type,
        resource=resource,
        id_type=id_type,
        id_value=id_value,
        kwargs=kwargs
    )

    return EntraIdCredentialsProvider(managed_identity_config, token_manager_config)


def create_from_service_principal(
        client_id: str,
        client_credential: Any,
        tenant_id: Optional[str] = None,
        scopes: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        token_kwargs: Optional[dict] = {},
        app_kwargs: Optional[dict] = {},
        token_manager_config: Optional[TokenManagerConfig] = TokenManagerConfig(
            DEFAULT_EXPIRATION_REFRESH_RATIO,
            DEFAULT_LOWER_REFRESH_BOUND_MILLIS,
            DEFAULT_TOKEN_REQUEST_EXECUTION_TIMEOUT_IN_MS,
            RetryPolicy(
                DEFAULT_MAX_ATTEMPTS,
                DEFAULT_DELAY_IN_MS
                )
            )) -> EntraIdCredentialsProvider:
    """
    Create a credential provider from a service principal.

    :param client_credential: Service principal credentials.
    :param client_id: Service principal client ID.
    :param scopes: Service principal scopes. Fallback to default scopes if None.
    :param timeout: Service principal timeout.
    :param tenant_id: Service principal tenant ID.
    :param token_kwargs: Optional token arguments to pass to service identity provider.
    :param app_kwargs: Optional keyword arguments to pass to service principal application.
    :param token_manager_config: Token manager specific configuration.
    :return: EntraIdCredentialsProvider instance.
    """
    service_principal_config = ServicePrincipalIdentityProviderConfig(
        client_credential=client_credential,
        client_id=client_id,
        scopes=scopes,
        timeout=timeout,
        tenant_id=tenant_id,
        app_kwargs=app_kwargs,
        token_kwargs=token_kwargs,
    )

    return EntraIdCredentialsProvider(service_principal_config, token_manager_config)


import pytest
from msal import TokenCache

from redis_entraid.identity_provider import EntraIDIdentityProvider


class TestEntraIDIdentityProvider:
    CUSTOM_CACHE = TokenCache()

    def test_request_token_from_service_principal_identity(self, identity_provider: EntraIDIdentityProvider):
        assert identity_provider.request_token(force_refresh=True)

    @pytest.mark.parametrize(
        "identity_provider",
        [
            {
                "idp_kwargs": {"token_cache": CUSTOM_CACHE},
            }
        ],
        indirect=True,
    )
    def test_request_token_caches_token_after_initial_request(self, identity_provider: EntraIDIdentityProvider):
        assert len(list(self.CUSTOM_CACHE.search(TokenCache.CredentialType.ACCESS_TOKEN))) == 0

        token = identity_provider.request_token()
        assert len(list(self.CUSTOM_CACHE.search(TokenCache.CredentialType.ACCESS_TOKEN))) == 1
        assert list(self.CUSTOM_CACHE.search(
            TokenCache.CredentialType.ACCESS_TOKEN
        ))[0].get('secret') == token.get_value()

        identity_provider.request_token()
        assert len(list(self.CUSTOM_CACHE.search(TokenCache.CredentialType.ACCESS_TOKEN))) == 1
        assert list(self.CUSTOM_CACHE.search(
            TokenCache.CredentialType.ACCESS_TOKEN
        ))[0].get('secret') == token.get_value()

        new_token = identity_provider.request_token(force_refresh=True)
        assert token.get_value() != new_token.get_value()
        assert len(list(self.CUSTOM_CACHE.search(TokenCache.CredentialType.ACCESS_TOKEN))) == 1
        assert list(self.CUSTOM_CACHE.search(
            TokenCache.CredentialType.ACCESS_TOKEN
        ))[0].get('secret') == new_token.get_value()



import asyncio
from time import sleep

import pytest
from redis.auth.token import TokenInterface

from redis_entraid.cred_provider import EntraIdCredentialsProvider
from redis_entraid.identity_provider import ManagedIdentityType
from tests.conftest import AuthType


class TestEntraIdCredentialsProvider:
    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "idp_kwargs": {"auth_type": AuthType.SERVICE_PRINCIPAL},
            },
        ],
        ids=["Service principal"],
        indirect=True,
    )
    def test_get_credentials(self, credential_provider: EntraIdCredentialsProvider):
        credentials = credential_provider.get_credentials()
        assert len(credentials) == 2


    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "idp_kwargs": {"auth_type": AuthType.MANAGED_IDENTITY},
            },
            {
                "idp_kwargs": {
                    "auth_type": AuthType.MANAGED_IDENTITY,
                    "identity_type": ManagedIdentityType.USER_ASSIGNED
                },
            }
        ],
        ids=["Managed Identity (System-assigned)", "Managed Identity (User-assigned)"],
        indirect=True,
    )
    @pytest.mark.managed_identity
    def test_get_credentials_managed_identity(self, credential_provider: EntraIdCredentialsProvider):
        credentials = credential_provider.get_credentials()
        assert len(credentials) == 2


    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"block_for_initial": False},
                "idp_kwargs": {"auth_type": AuthType.SERVICE_PRINCIPAL},
            },
        ],
        ids=["Service principal"],
        indirect=True,
    )
    @pytest.mark.asyncio
    async def test_get_credentials_async(self, credential_provider: EntraIdCredentialsProvider):
        credentials = await credential_provider.get_credentials_async()
        assert len(credentials) == 2


    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"block_for_initial": True},
                "idp_kwargs": {"auth_type": AuthType.MANAGED_IDENTITY},
            },
            {
                "idp_kwargs": {
                    "auth_type": AuthType.MANAGED_IDENTITY,
                    "identity_type": ManagedIdentityType.USER_ASSIGNED
                },
            }
        ],
        ids=["Managed Identity (System-assigned)", "Managed Identity (User-assigned)"],
        indirect=True,
    )
    @pytest.mark.asyncio
    @pytest.mark.managed_identity
    async def test_get_credentials_async_managed_identity(self, credential_provider: EntraIdCredentialsProvider):
        credentials = await credential_provider.get_credentials_async()
        assert len(credentials) == 2


    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"expiration_refresh_ratio": 0.00002},
            }
        ],
        indirect=True,
    )
    def test_get_credentials_executes_on_next(self, credential_provider: EntraIdCredentialsProvider):
        tokens = []

        def on_next(token: TokenInterface):
            nonlocal tokens
            tokens.append(token)

        credential_provider.on_next(on_next)

        # Run token manager
        credential_provider.get_credentials()
        sleep(1)

        assert len(tokens) > 0

    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"expiration_refresh_ratio": 0.00002},
            }
        ],
        indirect=True,
    )
    @pytest.mark.asyncio
    async def test_get_credentials_async_executes_on_next(self, credential_provider: EntraIdCredentialsProvider):
        tokens = []

        def on_next(token: TokenInterface):
            nonlocal tokens
            tokens.append(token)

        credential_provider.on_next(on_next)

        # Run token manager
        await credential_provider.get_credentials_async()
        await asyncio.sleep(1)

        assert len(tokens) > 0

    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"expiration_refresh_ratio": 0.00002},
            }
        ],
        indirect=True,
    )
    def test_get_credentials_executes_on_error(self, credential_provider: EntraIdCredentialsProvider):
        errors = []

        def on_next(token: TokenInterface):
            raise Exception("Some exception")

        def on_error(error: Exception):
            nonlocal errors
            errors.append(error)
            raise error

        credential_provider.on_next(on_next)
        credential_provider.on_error(on_error)

        # Run token manager
        credential_provider.get_credentials()
        sleep(1)

        assert len(errors) > 0
        assert str(errors[0]) == "Some exception"

    @pytest.mark.parametrize(
        "credential_provider",
        [
            {
                "cred_provider_kwargs": {"expiration_refresh_ratio": 0.00002},
            }
        ],
        indirect=True,
    )
    @pytest.mark.asyncio
    async def test_get_credentials_async_executes_on_error(self, credential_provider: EntraIdCredentialsProvider):
        errors = []

        async def on_next(token: TokenInterface):
            raise Exception("Some exception")

        async def on_error(error: Exception):
            nonlocal errors
            errors.append(error)
            raise error

        credential_provider.on_next(on_next)
        credential_provider.on_error(on_error)

        # Run token manager
        await credential_provider.get_credentials_async()
        await asyncio.sleep(1)

        assert len(errors) > 0
        assert str(errors[0]) == "Some exception"
